from yggdrasil.tests import generate_component_tests
from yggdrasil.communication.tests.test_FileComm import TestFileComm


generate_component_tests('file', TestFileComm, globals(), __file__,
                         class_attr='comm')

__all__ = []



def modelB_function2(in_val):
    # Only valid if in_val > 2
    out_val = 2 * in_val**2
    print("modelB_function2(%s) = %s" % (in_val, out_val))
    return in_val, out_val


def condition_function2(in_val):
    return (in_val > 2)



import os
import copy
import pprint
import yaml
from yggdrasil import tools
from yggdrasil.components import import_component
from yggdrasil.languages import get_language_ext
from yggdrasil.metaschema.datatypes import get_type_class
from yggdrasil.tests import long_running
from yggdrasil.examples import _example_dir
from yggdrasil.examples.tests import ExampleTstBase
from yggdrasil.metaschema.datatypes import encode_type


_all_lang = tools.get_supported_lang()
_full_lang = []
_typed_lang = []
for x in _all_lang:
    xdrv = import_component('model', x)
    if xdrv.full_language:
        _full_lang.append(x)
        if xdrv.is_typed:
            _typed_lang.append(x)
_full_lang = tuple(_full_lang)
_typed_lang = tuple(_typed_lang)


@long_running
class TestExampleTypes(ExampleTstBase):
    r"""Test the Types example."""

    example_name = 'types'
    iter_over = ['language', 'type', 'method']
    iter_skip = [(set(_typed_lang), '*', 'run_example_generic'),
                 (set(_typed_lang), '*', 'run_example_pointers'),
                 (set(['c']), '*', 'run_example_c_nolengths'),
                 (set(['c']), '*', 'run_example_c_prefixes'),
                 ('c', set(['string', 'bytes', 'unicode']), 'run_example_c_nolengths'),
                 ('c', set(['1darray', 'ndarray']), 'run_example_c_prefixes'),
                 ('*', set(['array']), 'run_example_split_array')]
    iter_flaky = [('c', 'instance', '*'), ('cpp', 'instance', '*')]
    iter_list_language = _full_lang
    iter_list_method = ['run_example', 'run_example_generic',
                        'run_example_pointers', 'run_example_c_nolengths',
                        'run_example_c_prefixes', 'run_example_split_array']

    def __init__(self, *args, **kwargs):
        self._output_files = None
        super(TestExampleTypes, self).__init__(*args, **kwargs)

    @property
    def datatype(self):
        r"""str: Name of datatype being used by the current iteration."""
        return self.iter_param.get('datatype', None)

    @property
    def output_files(self):
        r"""list: Output files for the run."""
        return self._output_files

    @classmethod
    def get_test_data(cls, typename):
        r"""Determine a test data set for the specified type.

        Args:
            typename (str): Name of datatype.

        Returns:
            object: Example of specified datatype.

        """
        typeclass = get_type_class(typename)
        testclass = typeclass.import_test_class()
        out = testclass._valid_decoded[0]
        return out

    @classmethod
    def check_received_data(cls, typename, x_recv):
        r"""Check that the received message is equivalent to the
        test data for the specified type.

        Args:
            typename (str): Name of datatype.
            x_recv (object): Received object.

        Raises:
            AssertionError: If the received message is not equivalent
                to the received message.

        """
        typeclass = get_type_class(typename)
        testclass = typeclass.import_test_class()
        x_sent = cls.get_test_data(typename)
        print('RECEIVED:')
        pprint.pprint(x_recv)
        print('EXPECTED:')
        pprint.pprint(x_sent)
        testclass.assert_result_equal(x_recv, x_sent)

    @classmethod
    def get_varstr(cls, vars_list, language, using_pointers=False,
                   length_prefix=False, dont_add_lengths=False):
        r"""Determine the vars string that should be used in the yaml.

        Args:
            vars_list (list): List of variables.
            language (str): Language being tested.
            using_pointers (bool, optional): If True and the tested
                language supports pointers, pointers will be used rather
                than explicit arrays. Defaults to False.
            length_prefix (bool, optional): If True, the length variables
                will be given prefixes instead of suffixes. Defaults to
                False.
            dont_add_lengths (bool, optional): If True, lengths will not
                be added to the definition or assignments. Defaults to
                False.

        Returns:
            str: Variable string.

        """
        out = []
        if length_prefix:
            length_fmt = 'length_%s'
            ndim_fmt = 'ndim_%s'
            shape_fmt = 'shape_%s'
        else:
            length_fmt = '%s_length'
            ndim_fmt = '%s_ndim'
            shape_fmt = '%s_shape'
        for v in vars_list:
            out.append(v['name'])
            typename = v['datatype']['type']
            if (language == 'c') and (not dont_add_lengths):
                if using_pointers:
                    if typename in ['string', 'bytes',
                                    'unicode', '1darray']:
                        out.append(length_fmt % v['name'])
                    elif typename in ['ndarray']:
                        out += [ndim_fmt % v['name'],
                                shape_fmt % v['name']]
                elif typename in ['string', 'bytes', 'unicode']:
                    out.append(length_fmt % v['name'])
        return ', '.join(out)
        
    @classmethod
    def setup_model(cls, language, typename, language_ext=None,
                    using_pointers=False, using_generics=False,
                    split_array=False, dont_add_lengths=False,
                    length_prefix=False, assign_kws=None, **kwargs):
        r"""Write the model file for the specified combination of
        language and type.

        Args:
            language (str): Language that model should be written in.
            typename (str): Type that should be expected by the model.
            language_ext (str, optional): Extension that should be used
                for the model file. If not provided, the extension is
                determined from the specified language.
            using_pointers (bool, optional): If True and the tested
                language supports pointers, pointers will be used rather
                than explicit arrays. Defaults to False.
            using_generics (bool, optional): If True and the tested
                language has a dedicated generic class, the generic
                type will be used rather than explict types. Defaults
                to False.
            split_array (bool, optional): If True and the tested datatype
                is an array, the variables will be split and specified
                explicitly in the yaml. Defaults to False.
            dont_add_lengths (bool, optional): If True, lengths will not
                be added to the definition or assignments. Defaults to
                False.
            length_prefix (bool, optional): If True, the length variables
                will be given prefixes instead of suffixes. Defaults to
                False.
            assign_kws (dict, optional): Keyword arguments for the calls
                to write_assign_to_output. Defaults to {}.
            **kwargs: Additional keyword arguments are passed to
                the write_function_def class method of the language
                driver.

        Returns:
            str: Full path to the file that was written.

        """
        if assign_kws is None:
            assign_kws = {}
        if language in ['c', 'c++', 'cpp']:
            # dont_add_lengths is only valid for C/C++
            kwargs['dont_add_lengths'] = dont_add_lengths
            kwargs['use_length_prefix'] = length_prefix
            assign_kws.setdefault('dont_add_lengths', dont_add_lengths)
            assign_kws.setdefault('use_length_prefix', length_prefix)
        yaml_fields = {'vars': False, 'dtype': False}
        if language_ext is None:
            language_ext = get_language_ext(language)
        modelfile = os.path.join(_example_dir, cls.example_name,
                                 'src', 'model' + language_ext)
        drv = import_component('model', language)
        if using_generics and drv.is_typed:
            testtype = {'type': 'any'}
        else:
            testdata = cls.get_test_data(typename)
            testtype = encode_type(testdata)
            using_generics = False
        if split_array and (typename == 'array'):
            inputs = [{'name': 'x%d' % i, 'datatype': x} for i, x in
                      enumerate(copy.deepcopy(testtype['items']))]
            outputs = [{'name': 'y%d' % i, 'datatype': x} for i, x in
                       enumerate(copy.deepcopy(testtype['items']))]
        else:
            inputs = [{'name': 'x',
                       'datatype': copy.deepcopy(testtype)}]
            outputs = [{'name': 'y',
                        'datatype': copy.deepcopy(testtype)}]
        # Write the model
        function_contents = []
        for i, o in zip(inputs, outputs):
            if using_pointers and drv.is_typed:
                for k in ['shape', 'length']:
                    i['datatype'].pop(k, None)
                    o['datatype'].pop(k, None)
            function_contents += drv.write_assign_to_output(
                o, i, outputs_in_inputs=drv.outputs_in_inputs,
                **assign_kws)
        lines = drv.write_function_def(
            'model', function_contents=function_contents,
            inputs=copy.deepcopy(inputs),
            outputs=copy.deepcopy(outputs),
            outputs_in_inputs=drv.outputs_in_inputs,
            opening_msg='IN MODEL', closing_msg='MODEL EXIT',
            print_inputs=True, print_outputs=True, **kwargs)
        with open(modelfile, 'w') as fd:
            print(modelfile)
            print('\n'.join(lines))
            fd.write('\n'.join(lines))
        os.environ['TEST_LANGUAGE'] = language
        os.environ['TEST_LANGUAGE_EXT'] = language_ext
        os.environ['TEST_TYPENAME'] = typename
        if language == 'c' and (not using_generics):
            yaml_fields['vars'] = True
            if typename in ['array', 'object']:
                yaml_fields['dtype'] = True
        if any(list(yaml_fields.values())):
            lines = []
            for io, io_vars in zip(['input', 'output'],
                                   [inputs, outputs]):
                lines += [io + 's:',
                          '  name: %s' % io]
                if yaml_fields['vars']:
                    lines.append(
                        '  vars: %s' % cls.get_varstr(
                            io_vars, language,
                            using_pointers=using_pointers,
                            length_prefix=length_prefix,
                            dont_add_lengths=dont_add_lengths))
                if yaml_fields['dtype']:
                    if len(io_vars) == 1:
                        dtype = io_vars[0]['datatype']
                    else:
                        dtype = {'type': 'array',
                                 'items': [x['datatype'] for
                                           x in io_vars]}
                    lines.append('  datatype:')
                    for x in yaml.dump(dtype).splitlines():
                        if "units: ''" in x:
                            continue
                        lines.append('    ' + x)
            os.environ['TEST_MODEL_IO'] = '\n    '.join(lines) + '\n'
        else:
            os.environ['TEST_MODEL_IO'] = ''
        return modelfile

    def check_results(self):
        r"""This should be overridden with checks for the result."""
        pass

    def run_example(self, **kwargs):
        r"""This runs an example in the correct language."""
        if self.yaml is not None:
            self._output_files = [self.setup_model(self.language,
                                                   self.datatype,
                                                   **kwargs)]
        super(TestExampleTypes, self).run_example()
        
    def run_example_generic(self):
        r"""Version of the example using generic type."""
        drv = import_component('model', self.language)
        if drv.is_typed:
            self.run_example(using_generics=True)

    def run_example_pointers(self):
        r"""Version of the example using pointers."""
        drv = import_component('model', self.language)
        if drv.is_typed:
            self.run_example(using_pointers=True)
    
    def run_example_c_nolengths(self):
        r"""Version of the example in C using pointers & no lengths."""
        drv = import_component('model', self.language)
        if ((drv.is_typed and (self.language == 'c')
             and (self.datatype in ['string', 'bytes', 'unicode']))):
            self.run_example(using_pointers=True,
                             dont_add_lengths=True)
    
    def run_example_c_prefixes(self):
        r"""Version of the example in C using pointers & prefixes."""
        drv = import_component('model', self.language)
        if ((drv.is_typed and (self.language == 'c')
             and (self.datatype in ['1darray', 'ndarray']))):
            self.run_example(using_pointers=True,
                             length_prefix=True)

    def run_example_split_array(self):
        r"""Version of the example where array is split."""
        if self.datatype == 'array':
            self.run_example(split_array=True)



import os
from pandas.testing import assert_frame_equal
from yggdrasil.components import create_component
from yggdrasil.examples.tests import ExampleTstBase


class TestExampleConditionalIO(ExampleTstBase):
    r"""Test the conditional_io example."""

    example_name = 'conditional_io'

    @property
    def expected_output_files(self):
        r"""list: Examples of expected output for the run."""
        return [os.path.join(self.yamldir, 'Output', 'output.txt')]
        
    @property
    def output_files(self):
        r"""list: Output files for the run."""
        return [os.path.join(self.yamldir, 'output.txt')]

    def read_file(self, fname):
        r"""Read in contents from a file.

        Args:
            fname (str): Full path to the file that should be read.

        Returns:
            object: File contents.

        """
        x = create_component('file', 'table', name='test',
                             address=fname, direction='recv',
                             as_array=True, recv_converter='pandas')
        msg = x.recv_array()[1]
        if msg is not None:
            msg = msg.sort_values(by=['InputMass']).reset_index(drop=True)
        x.close()
        return msg

    def assert_equal_file_contents(self, a, b):
        r"""Assert that the contents of two files are equivalent.

        Args:
            a (object): Contents of first file for comparison.
            b (object): Contents of second file for comparison.

        Raises:
            AssertionError: If the contents are not equal.

        """
        assert_frame_equal(a, b)

    def check_file_size(self, fname, fsize):
        r"""Check that file is the correct size.

        Args:
            fname (str): Full path to the file that should be checked.
            fsize (int): Size that the file should be in bytes.

        """
        pass



import unittest
import yggdrasil.drivers.tests.test_InterpretedModelDriver as parent


class TestSBMLModelParam(parent.TestInterpretedModelParam):
    r"""Test parameters for SBMLModelDriver class."""

    driver = 'SBMLModelDriver'

        
class TestSBMLModelDriverNoInit(TestSBMLModelParam,
                                parent.TestInterpretedModelDriverNoInit):
    r"""Test runner for SBMLModelDriver class without initing the driver."""
    
    def run_model_instance(self, **kwargs):  # pragma: sbml
        r"""Create a driver for a model and run it."""
        # This method of running dosn't work with SBML which requires io
        raise unittest.SkipTest("SBML requires I/O channels to run.")


class TestSBMLModelDriverNoStart(TestSBMLModelParam,
                                 parent.TestInterpretedModelDriverNoStart):
    r"""Test runner for SBMLModelDriver class without starting the driver."""
    pass


class TestSBMLModelDriver(TestSBMLModelParam,
                          parent.TestInterpretedModelDriver):
    r"""Test runner for SBMLModelDriver class."""
    pass



import copy
from collections import OrderedDict
from yggdrasil.metaschema.datatypes import (
    get_type_class, MetaschemaTypeError)
from yggdrasil.metaschema.datatypes.MetaschemaType import MetaschemaType
from yggdrasil.metaschema.properties.TypeMetaschemaProperty import (
    TypeMetaschemaProperty)


def create_multitype_class(types):
    r"""Create a MultiMetaschemaType class that wraps multiple
    classes.

    Args:
        types (list): List of names of types.

    Returns:
        class: Subclass of MultiMetaschemaType that add classes.

    """
    type_classes = OrderedDict()
    type_name = '_'.join(types)
    class_name = str('MultiMetaschemaType_%s' % type_name)
    for t in types:
        type_classes[t] = get_type_class(t)
    out = type(class_name, (MultiMetaschemaType, ),
               {'type_classes': type_classes,
                'name': type_name})
    return out


class MultiMetaschemaType(MetaschemaType):
    r"""Type class for handling behavior when more than one type is
    valid."""
    
    _dont_register = True
    inherit_properties = False
    type_classes = OrderedDict()

    def __init__(self, **typedef):
        typedef.setdefault('type', list(self.type_classes.keys()))
        self.type_instances = OrderedDict()
        super(MultiMetaschemaType, self).__init__(**typedef)
        
    @classmethod
    def get_type_class(cls, typedef=None, obj=None):
        r"""Get the type class from the provided typedef.

        """
        if (typedef is not None) and isinstance(typedef['type'],
                                                (str, bytes)):
            type_name = typedef['type']
        else:
            type_name = TypeMetaschemaProperty.encode(obj)
        if type_name not in cls.type_classes:
            raise MetaschemaTypeError(
                "Type '%s' not in set of supported types (%s)."
                % (type_name, list(cls.type_classes.keys())))
        return cls.type_classes[type_name]

    @classmethod
    def encode_data(cls, obj, typedef):
        r"""Encode an object's data.

        Args:
            obj (object): Object to encode.
            typedef (dict): Type definition that should be used to encode the
                object.

        Returns:
            string: Encoded object.

        """
        type_class = cls.get_type_class(typedef=typedef, obj=obj)
        return type_class.encode_data(obj, typedef)

    @classmethod
    def decode_data(cls, obj, typedef):
        r"""Decode an object.

        Args:
            obj (string): Encoded object to decode.
            typedef (dict): Type definition that should be used to decode the
                object.

        Returns:
            object: Decoded object.

        """
        type_class = cls.get_type_class(typedef=typedef, obj=obj)
        return type_class.decode_data(obj, typedef)

    @classmethod
    def transform_type(cls, obj, typedef=None):
        r"""Transform an object based on type info.

        Args:
            obj (object): Object to transform.
            typedef (dict): Type definition that should be used to transform the
                object.

        Returns:
            object: Transformed object.

        """
        type_class = cls.get_type_class(typedef=typedef, obj=obj)
        new_typedef = None
        if typedef is not None:
            new_typedef = dict(typedef, type=type_class.name)
        return type_class.transform_type(obj, typedef=new_typedef)

    @classmethod
    def coerce_type(cls, obj, typedef=None, **kwargs):
        r"""Coerce objects of specific types to match the data type.

        Args:
            obj (object): Object to be coerced.
            typedef (dict, optional): Type defintion that object should be
                coerced to. Defaults to None.
            **kwargs: Additional keyword arguments are metadata entries that may
                aid in coercing the type.

        Returns:
            object: Coerced object.

        """
        try:
            type_class = cls.get_type_class(typedef=typedef,
                                            obj=obj)
        except MetaschemaTypeError as e:
            raise ValueError(e)
        new_typedef = None
        if typedef is not None:
            new_typedef = dict(typedef, type=type_class.name)
        return type_class.coerce_type(obj, typedef=new_typedef, **kwargs)

    @classmethod
    def issubtype(cls, t):
        r"""Determine if this type is a subclass of the provided type.

        Args:
            t (str, list): Type name or list of type names to check against.

        Returns:
            bool: True if this type is a subtype of the specified type t.

        """
        if isinstance(t, list):
            return (len(set(cls.type_classes.keys()).intersection(t)) > 0)
        return (t in cls.type_classes)
    
    @classmethod
    def validate(cls, obj, raise_errors=False):
        r"""Validate an object to check if it could be of this type.

        Args:
            obj (object): Object to validate.
            raise_errors (bool, optional): If True, errors will be raised when
                the object fails to be validated. Defaults to False.

        Returns:
            bool: True if the object could be of this type, False otherwise.

        """
        for tcls in cls.type_classes.values():
            if tcls.validate(obj):
                return True
        if raise_errors:
            raise ValueError(("Object of type '%s' is not one of the accepted "
                              + "Python types for the allowed set or types (%s).") %
                             (type(obj), list(cls.type_classes.keys())))
        return False

    @classmethod
    def normalize(cls, obj):
        r"""Normalize an object, if possible, to conform to this type.

        Args:
            obj (object): Object to normalize.

        Returns:
            object: Normalized object.

        """
        for tcls in cls.type_classes.values():
            obj = tcls.normalize(obj)
        return obj

    @classmethod
    def encode_type(cls, obj, typedef=None, **kwargs):
        r"""Encode an object's type definition.

        Args:
            obj (object): Object to encode.
            typedef (dict, optional): Type properties that should be used to
                initialize the encoded type definition in certain cases.
                Defaults to None and is ignored.
            **kwargs: Additional keyword arguments are treated as additional
                schema properties.

        Returns:
            dict: Encoded type definition.

        """
        type_class = cls.get_type_class(typedef=typedef, obj=obj)
        new_typedef = None
        if typedef is not None:
            new_typedef = dict(typedef, type=type_class.name)
        return type_class.encode_type(obj, typedef=new_typedef, **kwargs)

    def update_typedef(self, **kwargs):
        r"""Update the current typedef with new values.

        Args:
            **kwargs: All keyword arguments are considered to be new type
                definitions. If they are a valid definition property, they
                will be copied to the typedef associated with the instance.

        Returns:
            dict: A dictionary of keyword arguments that were not added to the
                type definition.

        Raises:
            MetaschemaTypeError: If the current type does not match the type being
                updated to.

        """
        types = kwargs.pop('type', [])
        if self.type_instances:
            if set(types) != set(self.type_instances.keys()):
                raise MetaschemaTypeError(
                    "New types (%s) do not match old (%s)."
                    % (set(types), set(self.type_instances.keys())))
            for tcls in self.type_instances.values():
                tcls.update_typedef(**copy.deepcopy(kwargs))
        else:
            if set(types) != set(self.type_classes.keys()):
                raise MetaschemaTypeError(
                    "New types (%s) do not match class's (%s)."
                    % (set(types), set(self.type_classes.keys())))
            for t, tcls in self.type_classes.items():
                self.type_instances[t] = tcls(**copy.deepcopy(kwargs))
        return super(MultiMetaschemaType, self).update_typedef(
            type=types, **kwargs)

    @classmethod
    def definition_schema(cls):
        r"""JSON schema for validating a type definition schema."""
        types = list(cls.type_classes.keys())
        out = {'title': cls.name,
               'description': cls.description,
               'type': 'object',
               'required': copy.deepcopy(cls.definition_properties),
               'properties': {'type': {'oneOf': [
                   {'enum': types},
                   {'type': 'array',
                    'items': {'enum': types}}]}}}
        return out

    @classmethod
    def metadata_schema(cls):
        r"""JSON schema for validating a JSON serialization of the type."""
        types = list(cls.type_classes.keys())
        out = {'title': cls.name,
               'description': cls.description,
               'type': 'object',
               'required': copy.deepcopy(cls.metadata_properties),
               'properties': {'type': {'oneOf': [
                   {'enum': types},
                   {'type': 'array',
                    'items': {'enum': types}}]}}}
        return out

    @classmethod
    def _generate_data(cls, typedef):
        r"""Generate mock data for the specified type.

        Args:
            typedef (dict): Type definition.

        Returns:
            object: Python object of the specified type.

        """
        typedef = copy.deepcopy(typedef)
        types = typedef.pop('type', [])
        for t in types:
            try:
                return cls.type_classes[t].generate_data(
                    dict(typedef, type=t))
            except BaseException:
                pass
        raise NotImplementedError  # pragma: debug



import numpy as np
import pandas as pd
from collections import OrderedDict
import yggdrasil.drivers.tests.test_InterpretedModelDriver as parent


class TestRModelParam(parent.TestInterpretedModelParam):
    r"""Test parameters for RModelDriver."""

    driver = "RModelDriver"


class TestRModelDriverNoInit(TestRModelParam,
                             parent.TestInterpretedModelDriverNoInit):
    r"""Test runner for RModelDriver without init."""

    def test_is_library_installed(self):
        r"""Test is_library_installed for invalid library."""
        self.assert_equal(
            self.import_cls.is_library_installed('invalid_unicorn'),
            False)

    def test_python2language(self):
        r"""Test python2language."""
        test_vars = [(np.string_('hello'), 'hello'),
                     ((np.string_('hello'), ), ('hello', )),
                     ([np.string_('hello')], ['hello']),
                     ({np.string_('hello'): np.string_('hello')},
                      {'hello': 'hello'}),
                     (OrderedDict([(np.string_('hello'), np.string_('hello'))]),
                      OrderedDict([('hello', 'hello')]))]
        test_vars.append((
            pd.DataFrame.from_dict({'a': np.zeros(5, dtype='int64')}),
            pd.DataFrame.from_dict({'a': np.zeros(5, dtype='int32')})))
        for a, b in test_vars:
            self.assert_equal(self.import_cls.python2language(a), b)


class TestRModelDriverNoStart(TestRModelParam,
                              parent.TestInterpretedModelDriverNoStart):
    r"""Test runner for RModelDriver without start."""
    pass


class TestRModelDriver(TestRModelParam,
                       parent.TestInterpretedModelDriver):
    r"""Test runner for RModelDriver."""
    pass



import numpy as np
from yggdrasil.components import ComponentBase


class FilterBase(ComponentBase):
    r"""Base class for message filters.

    Args:
        initial_state (dict, optional): Dictionary of initial state variables
            that should be set when the filter is created.

    """

    _filtertype = None
    _schema_type = 'filter'
    _schema_subtype_key = 'filtertype'
    _schema_properties = {'initial_state': {'type': 'object'}}

    def __init__(self, *args, **kwargs):
        self._state = {}
        super(FilterBase, self).__init__(*args, **kwargs)
        if self.initial_state:
            self._state = self.initial_state

    def evaluate_filter(self, x):
        r"""Call filter on the provided message.

        Args:
            x (object): Message object to filter.

        Returns:
            bool: True if the message will pass through the filter, False otherwise.

        """
        raise NotImplementedError

    def __call__(self, x):
        r"""Call filter on the provided message.

        Args:
            x (object): Message object to filter.

        Returns:
            bool: True if the message will pass through the filter, False otherwise.

        """
        out = self.evaluate_filter(x)
        if isinstance(out, np.ndarray):
            assert(out.dtype == bool)
            out = bool(out.all())
        elif isinstance(out, np.bool_):
            out = bool(out)
        try:
            assert(isinstance(out, bool))
        except AssertionError:  # pragma: debug
            print(out, type(out))
            raise
        return out

    @classmethod
    def get_testing_options(cls):
        r"""Get testing options for the filter class.

        Returns:
            list: Mutiple dictionaries of keywords and messages that will
                pass/fail for those keywords.
        
        """
        return [{'error': [(1, NotImplementedError)],
                 'kwargs': {'initial_state': {'a': int(1)}}}]



import os
from yggdrasil.tests import assert_equal, assert_raises
from yggdrasil import config, tools


def test_YggConfigParser():
    r"""Ensure that get returns proper defaults etc."""
    x = config.YggConfigParser()
    x.add_section('test_section')
    x.set('test_section', 'test_option', 'test_value')
    assert_raises(RuntimeError, x.update_file)
    assert_equal(x.file_to_update, None)
    assert_equal(x.get('test_section', 'test_option'), 'test_value')
    assert_equal(x.get('test_section', 'fake_option'), None)
    assert_equal(x.get('test_section', 'fake_option', 5), 5)
    assert_equal(x.get('fake_section', 'fake_option'), None)
    assert_equal(x.get('fake_section', 'fake_option', 5), 5)


def test_update_language_config():
    r"""Test updating configuration for installed languages."""
    languages = tools.get_supported_lang()
    config.update_language_config(overwrite=True, verbose=True)
    try:
        config.update_language_config(
            languages[0], disable_languages=[languages[0]])
        config.update_language_config(
            languages[0], disable_languages=[languages[0]],
            enable_languages=[languages[0]])
    finally:
        config.update_language_config(
            languages[0], enable_languages=[languages[0]])


def test_cfg_logging():
    r"""Test cfg_logging."""
    lvl = config.get_ygg_loglevel()
    config.set_ygg_loglevel(lvl)
    os.environ['YGG_SUBPROCESS'] = 'True'
    lvl = config.get_ygg_loglevel()
    config.set_ygg_loglevel(lvl)
    config.cfg_logging()
    del os.environ['YGG_SUBPROCESS']



from yggdrasil.metaschema.properties.tests import (
    test_MetaschemaProperty as parent)
from yggdrasil.metaschema.properties.tests.test_ArgsMetaschemaProperty import (
    ValidArgsClass1, ValidArgsClass2, ValidArgsClass3,
    ValidArgsClass4, InvalidArgsClass)


class TestKwargsMetaschemaProperty(parent.TestMetaschemaProperty):
    r"""Test class for KwargsMetaschemaProperty class."""
    
    _mod = 'KwargsMetaschemaProperty'
    _cls = 'KwargsMetaschemaProperty'
    
    def __init__(self, *args, **kwargs):
        super(TestKwargsMetaschemaProperty, self).__init__(*args, **kwargs)
        self._valid = []
        self._invalid = []
        for cls in [ValidArgsClass1, ValidArgsClass2,
                    ValidArgsClass3, ValidArgsClass4]:
            cls_inst = cls(*(cls.test_args), **(cls.test_kwargs))
            self._valid.append((cls_inst, cls.valid_kwargs))
            self._invalid.append((cls_inst, cls.invalid_kwargs))
        valid_type = ValidArgsClass1.valid_kwargs
        invalid_type = ValidArgsClass1.invalid_kwargs
        self._encode_errors = [int(1), InvalidArgsClass]
        self._valid_compare = [(valid_type, valid_type)]
        self._invalid_compare = [(valid_type, invalid_type)]



import os
import sys
import importlib
from yggdrasil import platform, tools
from yggdrasil.metaschema.datatypes.MetaschemaType import MetaschemaType


class ClassMetaschemaType(MetaschemaType):
    r"""Type for evaluating classes."""

    name = 'class'
    description = 'Type for Python classes.'
    python_types = (type, )
    encoded_type = 'string'
    cross_language_support = False

    @classmethod
    def normalize(cls, obj, working_dir=None):
        r"""Normalize an object, if possible, to conform to this type.

        Args:
            obj (object): Object to normalize.
            working_dir (str, optional): Working directory that should
                be used to make relative paths absolute. Defaults to None.

        Returns:
            object: Normalized object.

        """
        if isinstance(obj, (str, bytes)):
            try:
                obj_str = tools.bytes2str(obj)
                obj = cls.decode_data(obj_str, {'type': cls.name},
                                      working_dir=working_dir)
            except (ValueError, AttributeError):
                pass
        return obj

    @classmethod
    def encode_data(cls, obj, typedef):
        r"""Encode an object's data.

        Args:
            obj (object): Object to encode.
            typedef (dict): Type definition that should be used to encode the
                object.

        Returns:
            string: Encoded object.

        """
        mod = obj.__module__
        fun = obj.__name__
        out = '%s:%s' % (mod, fun)
        return out

    @classmethod
    def decode_data(cls, obj, typedef, working_dir=None):
        r"""Decode an object.

        Args:
            obj (string): Encoded object to decode.
            typedef (dict): Type definition that should be used to decode the
                object.
            working_dir (str, optional): Working directory that should
                be used to make relative paths absolute. Defaults to None.

        Returns:
            object: Decoded object.

        """
        if not isinstance(obj, (str, bytes)):
            return obj
        pkg_mod = obj.split(':')
        if (len(pkg_mod) == 3) and platform._is_win:  # pragma: windows
            pkg_mod = [pkg_mod[0] + ':' + pkg_mod[1], pkg_mod[2]]
        if len(pkg_mod) != 2:
            raise ValueError("Could not parse %s string: %s"
                             % (cls.name, obj))
        mod, fun = pkg_mod[:]
        moddir, mod = os.path.split(mod)
        if working_dir is not None:
            moddir = os.path.normpath(os.path.join(working_dir, moddir))
        if mod.endswith('.py'):
            mod = os.path.splitext(mod)[0]
        try:
            modobj = importlib.import_module(mod)
        except ImportError:
            if not moddir:
                raise
            sys.path.append(os.path.abspath(moddir))
            modobj = importlib.import_module(mod)
        if not hasattr(modobj, fun):
            raise AttributeError("Module %s has no %s %s"
                                 % (modobj, cls.name, fun))
        return getattr(modobj, fun)

    @classmethod
    def _generate_data(cls, typedef):
        r"""Generate mock data for the specified type.

        Args:
            typedef (dict): Type definition.

        Returns:
            object: Python object of the specified type.

        """
        return ClassMetaschemaType



r"""This package provides a framework for integrating models across languages
such that they can be run simultaneously, passing input back and forth."""
from yggdrasil import platform
import os
import sys
import glob
import shutil
import logging
import argparse
import subprocess
import importlib
from ._version import get_versions
from yggdrasil import config
_test_package_name = None
_test_package = None
logging.basicConfig()
logger = logging.getLogger(__name__)
order = ['pytest', 'nose']
try:
    _test_package_name = order[0]
    _test_package = importlib.import_module(_test_package_name)
except ImportError:  # pragma: debug
    try:
        _test_package_name = order[1]
        _test_package = importlib.import_module(_test_package_name)
    except ImportError:
        _test_package_name = None
        _test_package = None
    

if platform._is_win:  # pragma: windows
    # This is required to fix crash on Windows in case of Ctrl+C
    # https://github.com/ContinuumIO/anaconda-issues/issues/905#issuecomment-232498034
    os.environ['FOR_DISABLE_CONSOLE_CTRL_HANDLER'] = 'T'


if not os.path.isfile(config.usr_config_file):  # pragma: no cover
    from yggdrasil.languages import install_languages
    shutil.copy(config.def_config_file, config.usr_config_file)
    install_languages.install_all_languages(from_setup=True)
    config.update_language_config()
    with open(config.usr_config_file, 'r') as fd:
        print(fd.read())


def expand_and_add(path, path_list, dir_list):  # pragma: no cover
    r"""Expand the specified path and add it's expanded forms to the provided
    list.

    Args:
        path (str): Absolute/relative path with or without regex wildcard
            expressions to expand.
        path_list (list): Existing list that expansions should be added to.
        dir_list (list): Directories that should be tried for relative paths.

    Returns:
        int: The number of expansions added to path_list.

    """
    if os.path.isabs(path):
        matches = sorted(glob.glob(path))
        path_list += matches
        return len(matches)
    # Try checking for function
    for func_sep in ['::', ':']:
        if (func_sep in path):
            prepath, mod = path.rsplit(func_sep, 1)
            nadded = expand_and_add(prepath, path_list, dir_list)
            if nadded:
                for i in range(-nadded, 0):
                    path_list[i] += '%s%s' % (func_sep, mod)
                return nadded
    # Try directory prefixes
    for path_prefix in dir_list:
        nadded = expand_and_add(os.path.join(path_prefix, path),
                                path_list, dir_list)
        if nadded:
            return nadded
    return 0


def run_tsts(**kwargs):  # pragma: no cover
    r"""Run tests for the package. Relative paths are interpreted to be
    relative to the package root directory.

    Args:
        verbose (bool, optional): If True, set option '-v' which
            increases the verbosity. Defaults to True.
        nocapture (bool, optional): If True, set option '--nocapture'
            ('--capture=no' with pytest) which allows messages to be printed to
            stdout. Defaults to True.
        stop (bool, optional): If True, set option '--stop' ('--exitfirst' for
            pytest) which stops tests at the first failure. Defaults to True.
        nologcapture (bool, optional): If True, set option '--nologcapture'
            which allows logged messages to be printed. Defaults to True.
        withcoverage (bool, optional): If True, set option '--with-coverage'
            which invokes coverage. Defaults to True.
        withexamples (bool, optional): If True, example testing will be
            enabled. Defaults to False.
        language (str, optional): Language to test. Defaults to None
            and all languages will be tested.

    """
    if '-h' not in sys.argv:
        if _test_package is None:
            raise RuntimeError("Could not locate test runner pytest or nose.")
        elif _test_package_name == 'pytest':
            test_cmd = 'pytest'
        elif _test_package_name == 'nose':
            test_cmd = 'nosetests'
        else:
            raise RuntimeError("Unsupported test package: '%s'"
                               % _test_package_name)
    parser = argparse.ArgumentParser(
        description='Run yggdrasil tests.')
    arguments = [
        (['withcoverage', 'with-coverage'], ['nocover', 'no-cover'],
         True, {'help': 'Record coverage during tests.'}),
        (['withexamples', 'with-examples'], ['noexamples', 'no-examples'],
         False, {'help': 'Run example tests when encountered.'}),
        (['longrunning', 'long-running'], ['nolongrunning', 'no-long-running'],
         False, {'help': 'Run long tests when encounterd.'}),
        (['verbose', 'v'], ['quiet'],
         True, {'help': ('Increase verbosity of output from '
                         'the test runner.')}),
        (['nocapture', 's'], ['capture'],
         True, {'help': 'Don\'t capture output from tests.'}),
        (['stop', 'x'], ['dontstop', 'dont-stop'],
         True, {'help': 'Stop after first test failure.'}),
        (['nologcapture'], ['logcapture'],
         True, {'help': ('Don\'t capture output from log '
                         'messages generated during tests.')}),
        (['validatecomponents', 'validate-components'],
         ['skipcomponentvalidation', 'skip-component-validation'],
         False,
         {'help': ('Validate components on creation. This causes '
                   'a decrease in performance so it is turned off '
                   'by default.')}),
        (['noflaky', 'no-flaky'], ['flaky'],
         False, {'help': 'Don\'t re-run flaky tests.'}),
        (['debug'], ['nodebug'],
         False, {'help': 'Turn on debug messages.'})]
    for pos_dest, neg_dest, default, kws in arguments:
        dest = pos_dest[0]
        for x in [pos_dest, neg_dest]:
            for i, y in enumerate(x):
                if len(y) == 1:
                    x[i] = '-' + y
                else:
                    x[i] = '--' + y
        if kwargs.get(dest, default):
            if kws['help'].startswith('Don\'t'):
                kws['help'].split('Don\'t', 1)[-1]
                kws['help'] = kws['help'].replace(
                    kws['help'][0], kws['help'][0].upper(), 1)
            else:
                kws['help'] = kws['help'].replace(
                    kws['help'][0], kws['help'][0].lower(), 1)
                kws['help'] = 'Don\'t ' + kws['help']
            parser.add_argument(*neg_dest, action='store_false',
                                dest=dest, **kws)
        else:
            parser.add_argument(*pos_dest, action='store_true',
                                dest=dest, **kws)
    parser.add_argument('--language', '--languages', default=[],
                        nargs="+", type=str,
                        help='Language(s) that should be tested.')
    parser.add_argument('--default-comm', '--defaultcomm', type=str,
                        help=('Comm type that default should be set '
                              'to before running tests.'))
    parser.add_argument('--ci', action='store_true',
                        help=('Perform addition operations required '
                              'for testing on continuous integration '
                              'services.'))
    suite_args = ('--test-suite', '--test-suites')
    suite_kws = dict(nargs='+', action="extend", type=str,
                     choices=['examples', 'examples_part1',
                              'examples_part2', 'types', 'timing'],
                     help='Test suite(s) that should be run.',
                     dest='test_suites')
    try:
        parser.add_argument(*suite_args, **suite_kws)
    except ValueError:
        # 'extend' introduced in 3.8
        suite_kws['action'] = 'append'
        suite_kws.pop('nargs')
        parser.add_argument(*suite_args, **suite_kws)
    args, extra_argv = parser.parse_known_args()
    initial_dir = os.getcwd()
    package_dir = os.path.dirname(os.path.abspath(__file__))
    error_code = 0
    # Peform ci tests/operations
    # Call bash script?
    if args.ci:
        extra_argv += ['-c', 'setup.cfg', '--cov-config=.coveragerc']
    # Separate out paths from options
    argv = [test_cmd]
    test_paths = []
    opt_val = 0
    for x in extra_argv:
        if opt_val > 0:
            argv.append(x)
            opt_val -= 1
        elif x.endswith('yggtest'):
            pass
        elif x.startswith('-'):
            argv.append(x)
            if (_test_package_name == 'pytest') and (x in ['-c']):
                opt_val = 1
        else:
            test_paths.append(x)
    if args.test_suites:
        for x in args.test_suites:
            if x == 'examples':
                args.withexamples = True
                test_paths.append('examples')
            elif x == 'examples_part1':
                args.withexamples = True
                test_paths.append(os.path.join(
                    'examples', 'tests', 'test_[a-g]*.py'))
            elif x == 'examples_part2':
                args.withexamples = True
                test_paths.append(os.path.join(
                    'examples', 'tests', 'test_[g-z]*.py'))
            # elif x.startswith('examples_'):
            #     args.withexamples = True
            #     test_paths.append(os.path.join(
            #         'examples', 'tests',
            #         'test_%s*.py'.format(x.split('examples_')[-1])))
            elif x == 'types':
                args.withexamples = True
                args.longrunning = True
                test_paths.append(os.path.join('examples', 'tests',
                                               'test_types.py'))
            elif x == 'timing':
                args.longrunning = True
                test_paths.append(os.path.join('tests', 'test_timing.py'))
    if _test_package_name == 'nose':
        argv += ['--detailed-errors', '--exe']
    if args.verbose:
        argv.append('-v')
    if args.nocapture:
        argv.append('-s')
    if args.stop:
        argv.append('-x')
    if args.nologcapture and (_test_package_name == 'nose'):
        argv.append('--nologcapture')
    if args.withcoverage:
        if _test_package_name == 'nose':
            argv.append('--with-coverage')
            argv.append('--cover-package=yggdrasil')
        elif _test_package_name == 'pytest':
            # See information about getting coverage of test fixtures
            # https://pytest-cov.readthedocs.io/en/stable/plugins.html
            argv.append('--cov=%s' % package_dir)
            # argv.append('--cov-append')
    if args.noflaky:
        if _test_package_name == 'pytest':
            argv += ['-p', 'no:flaky']
    else:
        if _test_package_name == 'nose':
            argv.append('--with-flaky')
    # Get expanded tests to allow for paths that are relative to either the
    # yggdrasil root directory or the current working directory
    expanded_test_paths = []
    if not test_paths:
        expanded_test_paths.append(package_dir)
    else:
        for x in test_paths:
            if not expand_and_add(x, expanded_test_paths,
                                  [package_dir, os.getcwd()]):
                expanded_test_paths.append(x)
    argv += expanded_test_paths
    # Run test command and perform cleanup before logging any errors
    logger.info("Running %s from %s", argv, os.getcwd())
    new_config = {}
    old_config = {}
    new_env = {}
    old_env = {}
    pth_file = 'ygg_coverage.pth'
    assert(not os.path.isfile(pth_file))
    try:
        # Set env
        if args.withexamples:
            new_env['YGG_ENABLE_EXAMPLE_TESTS'] = 'True'
        if args.language:
            from yggdrasil.components import import_component
            args.language = [import_component('model', x).language
                             for x in args.language]
            new_env['YGG_TEST_LANGUAGE'] = ','.join(args.language)
        if args.default_comm:
            new_env['YGG_DEFAULT_COMM'] = args.default_comm
        if args.longrunning:
            new_env['YGG_ENABLE_LONG_TESTS'] = 'True'
        if args.withcoverage:
            new_env['COVERAGE_PROCESS_START'] = 'True'
            # if _test_package_name == 'pytest':
            #     # See information about getting coverage of test fixtures
            #     # https://pytest-cov.readthedocs.io/en/stable/plugins.html
            #     new_env['COV_CORE_SOURCE'] = package_dir
            #     new_env['COV_CORE_CONFIG'] = '.coveragerc'
            #     new_env['COV_CORE_DATAFILE'] = '.coverage.eager'
            with open(pth_file, 'w') as fd:
                fd.write("import coverage; coverage.process_startup()")
        if args.debug:
            new_config[('debug', 'ygg')] = 'DEBUG'
            new_config[('debug', 'client')] = 'DEBUG'
        if args.test_suites and ('timing' in args.test_suites):
            new_env['YGG_TEST_PRODUCTION_RUNS'] = 'True'
        if not args.validatecomponents:
            new_env['YGG_SKIP_COMPONENT_VALIDATION'] = 'True'
        # Update environment and config
        for k, v in new_env.items():
            old_env[k] = os.environ.get(k, None)
            os.environ[k] = v
        for k, v in new_config.items():
            old_config[k] = config.ygg_cfg_usr.get(k[0], k[1], None)
            config.ygg_cfg_usr.set(k[0], k[1], v)
        config.ygg_cfg_usr.update_file()
        # Perform CI specific pretest operations
        if args.ci:
            top_dir = os.path.dirname(os.getcwd())
            src_cmd = ('python -c \"import versioneer; '
                       'print(versioneer.get_version())\"')
            dst_cmd = ('python -c \"import yggdrasil; '
                       'print(yggdrasil.__version__)\"')
            src_ver = subprocess.check_output(src_cmd, shell=True)
            dst_ver = subprocess.check_output(dst_cmd, shell=True,
                                              cwd=top_dir)
            if src_ver != dst_ver:  # pragma: debug
                raise RuntimeError(("Versions do not match:\n"
                                    "\tSource version: %s\n"
                                    "\tBuild  version: %s\n")
                                   % (src_ver, dst_ver))
            if os.environ.get("INSTALLR", None) == "1":
                from yggdrasil import tools
                print(tools.which("R"))
                print(tools.which("Rscript"))
            subprocess.check_call(["flake8", "yggdrasil"])
            if os.environ.get("YGG_CONDA", None):
                subprocess.check_call(["python", "create_coveragerc.py"])
            if not os.path.isfile(".coveragerc"):
                raise RuntimeError(".coveragerc file dosn't exist.")
            with open(".coveragerc", "r") as fd:
                print(fd.read())
            subprocess.check_call(["ygginfo", "--verbose"])
        error_code = subprocess.call(argv)
    except BaseException:
        logger.exception('Error in running test.')
        error_code = -1
    finally:
        os.chdir(initial_dir)
        for k, v in old_env.items():
            if v is None:
                os.environ.pop(k, None)
            else:
                os.environ[k] = v
        for k, v in old_config.items():
            if v is None:
                config.ygg_cfg_usr.pop(k[0], k[1])
            else:
                config.ygg_cfg_usr.set(k[0], k[1], v)
        config.ygg_cfg_usr.update_file()
        if os.path.isfile(pth_file):
            os.remove(pth_file)
    return error_code


__all__ = []
__version__ = get_versions()['version']
del get_versions



from yggdrasil.communication.filters.FilterBase import FilterBase


class DirectFilter(FilterBase):
    r"""Class that always passes messages."""

    _filtertype = 'direct'

    def evaluate_filter(self, x):
        r"""Call filter on the provided message.

        Args:
            x (object): Message object to filter.

        Returns:
            bool: True if the message will pass through the filter, False otherwise.

        """
        return True

    @classmethod
    def get_testing_options(cls):
        r"""Get testing options for the filter class.

        Returns:
            list: Mutiple dictionaries of keywords and messages that will
                pass/fail for those keywords.
        
        """
        return [{'pass': [1, 0]}]



import os
import argparse
from yggdrasil.examples.tests.test_transforms import TestExampleTransforms
from yggdrasil.languages import get_language_ext
from yggdrasil.runner import run
_this_dir = os.path.dirname(__file__)


def main(language, transform, language_ext=None):
    yamlfile = os.path.join(_this_dir, 'transforms.yml')
    modelfile, env = TestExampleTransforms.setup_model(
        language, transform, language_ext=language_ext)
    os.environ.update(env)
    try:
        run(yamlfile)
    finally:
        if os.path.isfile(modelfile):
            os.remove(modelfile)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        "Test application of a transform for the specified language.")
    # TODO: Set to list of installed languages/transforms?
    parser.add_argument('language',
                        help='Language that should be tested.')
    parser.add_argument('transform',
                        help='Name of data transform that should be tested.')
    args = parser.parse_args()
    args.language_ext = get_language_ext(args.language)
    main(args.language, args.transform, args.language_ext)



import numpy as np
import copy
import pandas
from yggdrasil.communication.transforms.TransformBase import TransformBase
from yggdrasil.metaschema.datatypes import type2numpy
from yggdrasil.serialize import (
    consolidate_array, pandas2numpy, numpy2pandas, dict2list)
from yggdrasil.metaschema.properties.ScalarMetaschemaProperties import (
    _valid_types, _flexible_types)


class ArrayTransform(TransformBase):
    r"""Class for consolidating values into an array."""
    _transformtype = 'array'
    _schema_properties = {'field_names': {'type': 'array',
                                          'items': {'type': 'string'}}}

    def set_original_datatype(self, datatype):
        r"""Set datatype.

        Args:
            datatype (dict): Datatype.

        """
        super(ArrayTransform, self).set_original_datatype(datatype)
        if not self.field_names:
            self.field_names = self.original_datatype.get('field_names', None)
        if not self.field_names:
            if (((datatype['type'] == 'array')
                 and isinstance(datatype['items'], list)
                 and all([('title' in x) for x in
                          self.original_datatype['items']]))):
                self.field_names = [x.get('title', 'f%d' % i) for i, x in
                                    enumerate(self.original_datatype['items'])]
            elif datatype['type'] == 'object':
                self.field_names = list(datatype['properties'].keys())

    @classmethod
    def get_summary(cls, x, subtype=False):
        r"""Get subset of information summarizing an array element
        that can be used for comparison with other elements in the
        same row/column.

        Args:
            x (dict): Type definition for an array element.
            subtype (bool, optional): If True, the subtype, shape,
                and title information is included. Defaults to False.

        Returns:
            dict: Information about the array element.

        Raises:
            AssertionError: If x is not a valid type defintion for an
                array element.

        """
        if x['type'] == 'ndarray':
            s = x.get('shape', None)
            t = 'ndarray'
        elif x['type'] == '1darray':
            s = x.get('length', None)
            if s is not None:
                s = (s,)
            t = '1darray'
        elif ((x['type'] == 'scalar')
              or (x['type'] in _valid_types)):
            s = (1,)
            t = 'scalar'
        else:
            raise AssertionError(("Cannot convert elements of type '%s' "
                                  "to array elements.") % x['type'])
        subt = x.get('subtype', x['type'])
        title = x.get('title', None)
        assert(subt in _valid_types)
        if subtype:
            out = {'type': t, 'subtype': subt,
                   'shape': s, 'title': title}
            if subt not in _flexible_types:
                out['precision'] = x.get('precision', 0)
        else:
            out = {'type': t, 'shape': s}
        return out

    @classmethod
    def check_summary(cls, a, aidx, b, bidx):
        r"""Determine if two summary structures are equivalent,
        printing differences in the error if they are not.

        Args:
            a (dict): Summary information for an element type defintion.
            aidx (int): Index of element summarized by a that is used
                in the error message.
            b (dict): Summary information for an element type defintion.
            bidx (int): Index of element summarized by b that is used
                in the error message.

        Raises:
            AssertionError: If a and b are not equivalent.

        """
        if a == b:
            return
        assert(len(a) == len(b))
        err_msg = []
        for k in a.keys():
            if a[k] != b[k]:
                err_msg.append(("The %s of element %d (%s) dosn't "
                                "match element %d (%s=%s)")
                               % (k, aidx, a[k], bidx, k, b[k]))
        raise AssertionError('\n'.join(err_msg))

    @classmethod
    def check_element(cls, items, subtype=False):
        r"""Check that all elements in set of elements (e.g. row or
        column) are consistent.

        Args:
            items (list): Set of element type definitions.
            subtype (bool, optional): If True, subtype, precision, and
                title information are used in the comparison. Defaults
                to False. subtype should be True if checking column
                elements and False if checking row elements.

        Raises:
            AssertionError: If any elements are not consistent.

        """
        base_summary = cls.get_summary(items[0], subtype=subtype)
        for i, x in zip(range(1, len(items)), items[1:]):
            x_summary = cls.get_summary(x, subtype=subtype)
            cls.check_summary(x_summary, i, base_summary, 0)

    @classmethod
    def check_array_items(cls, items, order=None, items_as_columns=None):
        r"""Check that items are valid types for array columns.

        Args:
            items (list): Type definitions for elements.
            order (list, optional): Order that properties should be
                compared in for object schemas. Defaults to None and
                will be set based on the order of the keys in the
                first element (non-deterministic for Python 2.7).
            items_as_columns (bool, optional): If True, the items will
                be parsed under the assumption that each item contains
                the schema describing a column, possible as an array
                of elements. If None and the initial check fails
                when assuming items are rows, columns will be tried.
                Defaults to None.

        Raises:
            AssertionError: If the items are not valid.

        """
        if isinstance(items, dict):
            items = [items]
        assert(isinstance(items, (list, tuple)))
        if items[0]['type'] == 'array':
            base_types = items[0]['items']
            assert(isinstance(base_types, list))
        elif items[0]['type'] == 'object':
            if order is None:
                order = list(items[0]['properties'].keys())
            base_types = [items[0]['properties'][k] for k in order]
        elif items[0]['type'] in ['1darray', 'ndarray']:
            cls.check_element(items)
            return
        else:
            raise AssertionError("Per-element types of '%s' not supported."
                                 % items[0]['type'])
        try:
            cls.check_element(base_types, subtype=items_as_columns)
            base_summary = [cls.get_summary(x, subtype=(not items_as_columns))
                            for x in base_types]
            for i, x in zip(range(1, len(items)), items[1:]):
                assert(x['type'] == items[0]['type'])
                if x['type'] == 'array':
                    x_types = x['items']
                else:
                    x_types = [x['properties'][k] for k in order]
                assert(len(x_types) == len(base_types))
                if items_as_columns:
                    cls.check_element(x_types, subtype=True)
                x_summary = [cls.get_summary(t, subtype=(not items_as_columns))
                             for t in x_types]
                for ix, ibase in zip(x_summary, base_summary):
                    cls.check_summary(x_summary, i, base_summary, 0)
        except BaseException as e:
            if (((items_as_columns is None)
                 and all([(x['type'] == 'array') for x in items]))):
                try:
                    cls.check_array_items(items, order=order,
                                          items_as_columns=True)
                    return
                except BaseException:
                    pass
            raise e

    def validate_datatype(self, datatype):
        r"""Assert that the provided datatype is valid for this transformation.
        
        Args:
            datatype (dict): Datatype to validate.

        Raises:
            AssertionError: If the datatype is not valid.

        """
        if datatype['type'] in ['1darray', 'ndarray']:
            pass
        elif datatype['type'] == 'array':
            self.check_array_items(datatype['items'],
                                   order=self.field_names)
        elif datatype['type'] == 'object':
            order = self.field_names
            if order is None:
                order = list(datatype['properties'].keys())
            self.check_array_items([datatype['properties'][k]
                                    for k in order])
        else:
            raise AssertionError("Invalid datatypes: %s" % datatype)

    @classmethod
    def transform_array_items(cls, items, order=None):
        r"""Transform elements in an array.

        Args:
            items (list): Set of type definitions for array rows or
                columns that should be transformed into type
                definitions for a set of array columns.
            order (list, optional): Order in which properties should
                be added as columns for object type defintions. Defaults
                to None if not provided and the first object element will
                be used to get the order (non-deterministic on Python 2.7).

        Returns:
            list: Transformed array column type definitions.

        """
        if isinstance(items, dict):
            items = [items]
        assert(isinstance(items, (list, tuple)))
        if items[0]['type'] == 'array':
            base_types = items[0]['items']
            assert(isinstance(base_types, list))
        elif items[0]['type'] == 'object':
            if order is None:
                order = list(items[0]['properties'].keys())
            items = [dict(x, items=[dict(x['properties'][k], title=k)
                                    for k in order])
                     for x in items]
            base_types = items[0]['items']
        elif items[0]['type'] in ['1darray', 'ndarray']:
            return items
        base_summary = [cls.get_summary(x, subtype=True)
                        for x in base_types]
        if not all([(base_summary == [cls.get_summary(t, subtype=True)
                                      for t in x['items']])
                    for x in items[1:]]):
            items = [{'items': [copy.deepcopy(items[j]['items'][i])
                                for j in range(len(items))]}
                     for i in range(len(items[0]['items']))]
            base_types = items[0]['items']
        out = [dict(x, type='1darray',
                    subtype=x.get('subtype', x['type']))
               for x in base_types]
        for i, x in enumerate(out):
            if x['subtype'] in _flexible_types:
                x['precision'] = max(
                    [y['items'][i].get('precision', 0) for y in items])
                if x['precision'] == 0:
                    x.pop('precision')
        return out
        
    def transform_datatype(self, datatype):
        r"""Determine the datatype that will result from applying the transform
        to the supplied datatype.

        Args:
            datatype (dict): Datatype to transform.

        Returns:
            dict: Transformed datatype.

        """
        out = copy.deepcopy(datatype)
        if datatype['type'] == 'array':
            out['items'] = self.transform_array_items(
                out['items'], order=self.field_names)
        elif datatype['type'] == 'object':
            order = self.field_names
            if order is None:
                order = list(out['properties'].keys())
            out['type'] = 'array'
            out['items'] = self.transform_array_items(
                [dict(out['properties'][k], title=k)
                 for k in order])
            out.pop('properties', None)
        if self.field_names is not None:
            assert(len(self.field_names) == len(out['items']))
            for x, n in zip(out['items'], self.field_names):
                x['title'] = n
        return out
    
    def evaluate_transform(self, x, no_copy=False):
        r"""Call transform on the provided message.

        Args:
            x (object): Message object to transform.
            no_copy (bool, optional): If True, the transformation occurs in
                place. Otherwise a copy is created and transformed. Defaults
                to False.

        Returns:
            object: The transformed message.

        """
        out = x
        np_dtype = type2numpy(self.transformed_datatype)
        if isinstance(x, pandas.DataFrame):
            out = pandas2numpy(x).astype(np_dtype, copy=True)
        elif isinstance(x, np.ndarray):
            out = x.astype(np_dtype, copy=True)
        elif np_dtype and isinstance(x, (list, tuple, dict,
                                         np.ndarray)):
            if len(x) == 0:
                out = np.zeros(0, np_dtype)
            else:
                if isinstance(x, dict):
                    x = dict2list(x, order=np_dtype.names)
                out = consolidate_array(x, dtype=np_dtype)
        else:
            # warning?
            raise TypeError(("Cannot consolidate object of type %s "
                             "into a structured numpy array.") % type(x))
        if not no_copy:
            out = copy.deepcopy(out)
        return out
    
    @classmethod
    def get_testing_options(cls):
        r"""Get testing options for the transform class.

        Returns:
            list: Multiple dictionaries of keywords and messages before/after
                pairs that will result from the transform created by the provided
                keywords.
        
        """
        length = 5
        dtype = np.dtype([('f%d' % i, f) for i, f in enumerate(
            ['S5', 'i8', 'f8', 'c16'])])
        dtype_alt = np.dtype([('alt%d' % i, f) for i, f in enumerate(
            ['S5', 'i8', 'f8', 'c16'])])
        t = {'type': 'array',
             'items': [
                 {'type': '1darray', 'subtype': 'bytes',
                  'precision': 40, 'length': length},
                 {'type': '1darray', 'subtype': 'int',
                  'precision': 64, 'length': length},
                 {'type': '1darray', 'subtype': 'float',
                  'precision': 64, 'length': length},
                 {'type': '1darray', 'subtype': 'complex',
                  'precision': 128, 'length': length}]}
        t_prec = {
            'type': 'array',
            'items': [
                {'type': '1darray', 'subtype': 'bytes',
                 'length': length},
                {'type': '1darray', 'subtype': 'int',
                 'precision': 64, 'length': length},
                {'type': '1darray', 'subtype': 'float',
                 'precision': 64, 'length': length},
                {'type': '1darray', 'subtype': 'complex',
                 'precision': 128, 'length': length}]}
        t_arr = {'type': 'array',
                 'items': [{'type': 'array',
                            'items': [dict(i, type='scalar') for
                                      i in t['items']]}
                           for _ in range(length)]}
        t_arr_err = copy.deepcopy(t_arr)
        t_arr_err['items'][0]['items'][0]['type'] = 'null'
        t_obj = {'type': 'array',
                 'items': [{'type': 'object',
                            'properties': {
                                dtype_alt.names[i]: dict(t['items'][i],
                                                         type='scalar')
                                for i in range(len(t['items']))}}
                           for _ in range(length)]}
        t_arr_T = {
            'type': 'array',
            'items': [{'type': 'array',
                       'items': [dict(t['items'][i], type='scalar')
                                 for _ in range(length)]}
                      for i in range(len(t['items']))]}
        t_arr_prec = {
            'type': 'array',
            'items': [{'type': 'array',
                       'items': [dict(i, type='scalar') for
                                 i in t_prec['items']]}
                      for _ in range(length)]}
        t_alt = {'type': 'array',
                 'items': [dict(x, title=dtype_alt.names[i])
                           for i, x in enumerate(t['items'])]}
        x = np.zeros(length, dtype=dtype)
        x[dtype.names[0]][0] = b'hello'
        y = [x[n] for n in dtype.names]
        x2 = np.zeros((length, length), dtype=dtype)
        # y2 = [x2[n] for n in dtype2.names]
        return [{'kwargs': {'original_datatype': t},
                 'in/out': [(y, x),
                            ([], np.zeros(0, dtype=dtype))],
                 'in/out_t': [(t, t),
                              (t_arr_prec, t_prec),
                              (t_arr_T, t),
                              (t_obj, t_alt),
                              ({'type': 'null'}, AssertionError),
                              (t['items'][0], t['items'][0]),
                              ({'type': 'array',
                                'items': [dict(v, length=i)
                                          for i, v in enumerate(t['items'])]},
                               AssertionError),
                              (t_arr_err, AssertionError)]},
                {'in/out': [(x, x)]},
                {'in/out': [(x2, x2)]},
                {'kwargs': {'field_names': dtype_alt.names},
                 'in/out': [(x, x.astype(dtype_alt, copy=True))],
                 'in/out_t': [(t, t_alt)]},
                {'kwargs': {'original_datatype': t},
                 'in/out': [(numpy2pandas(x), x)],
                 'in/out_t': [(t, t)]},
                {'kwargs': {'original_datatype': t_arr},
                 'in/out': [(x.tolist(), x)],
                 'in/out_t': [(t_arr, t),
                              ({'type': 'array',
                                'items': t_arr['items'][0]}, t)]},
                {'in/out': [({n: x[n] for n in dtype.names}, x)],
                 'in/out_t': [({'type': 'object',
                                'properties': {n: i for n, i in
                                               zip(dtype.names, t['items'])}},
                               {'type': 'array',
                                'items': [dict(i, title=n) for n, i in
                                          zip(dtype.names, t['items'])]})]},
                {'kwargs': {'original_datatype': t_arr},
                 'in/out': [(x.tolist(), x)],
                 'in/out_t': [(t_arr, t)]},
                {'kwargs': {'original_datatype': t},
                 'in/out': [(None, TypeError)]},
                {'kwargs': {},
                 'in/out': [([0, 1, 2], AssertionError)]}]



"""
This module imports the configuration for yggdrasil.

.. todo::
   Remove reference to environment variables for accessing config options.

"""
import os
import sys
import json
import shutil
import logging
import warnings
import configparser
from collections import OrderedDict
from yggdrasil import tools
env_prefixes = tools.get_env_prefixes()
config_file = '.yggdrasil.cfg'
def_config_file = os.path.join(os.path.dirname(__file__), 'defaults.cfg')
usr_dir = os.path.expanduser('~')
if env_prefixes:
    usr_dir = env_prefixes[-1]
usr_config_file = os.path.join(usr_dir, config_file)
loc_config_file = os.path.join(os.getcwd(), config_file)
logger = logging.getLogger(__name__)


class YggConfigParser(configparser.ConfigParser, object):
    r"""Config parser that returns None if option not provided on get."""

    def __init__(self, files=None):
        self.files = files
        super(YggConfigParser, self).__init__()

    def reload(self):
        r"""Reload parameters from the original files."""
        for s in self.sections():
            self.remove_section(s)
        self._sections = self._dict()
        if self.files is not None:
            self.read(self.files)

    @property
    def file_to_update(self):
        r"""str: Full path to file that should be updated if update_file is
        called without an explicit file path."""
        out = None
        if self.files is not None:
            out = self.files[-1]
        return out

    def update_file(self, fname=None):
        r"""Write out updated contents to a file.

        Args:
            fname (str, optional): Full path to file where contents should be
               saved. If None, file_to_update is used. Defaults to None.

        Raises:
            RuntimeError: If fname is None and file_to_update is None.

        """
        if fname is None:
            fname = self.file_to_update
        if fname is None:
            raise RuntimeError("No file provided or set at creation.")
        with open(fname, 'w') as fd:
            self.write(fd)

    def read(self, *args, **kwargs):
        out = super(YggConfigParser, self).read(*args, **kwargs)
        alias_map = [(('debug', 'psi'), ('debug', 'ygg')),
                     (('debug', 'cis'), ('debug', 'ygg'))]
        for old, new in alias_map:
            v = self.get(*old)
            if v:  # pragma: debug
                self.set(new[0], new[1], v)
        return out

    @classmethod
    def from_files(cls, files, **kwargs):
        r"""Construct a config parser from a set of files.

        Args:
            files (list): One or more files that options should be read from in
                the order they should be loaded.
            **kwargs: Additional keyword arguments are passed to the class
                constructor.

        Returns:
           YggConfigParser: Config parser with information loaded from the
               provided files.

        """
        out = cls(files=files, **kwargs)
        out.reload()
        return out

    def set(self, section, option, value=None):
        """Set an option."""
        if not isinstance(value, str):
            value = json.dumps(value)
        super(YggConfigParser, self).set(section, option, value=value)

    def backwards_str2val(self, val):  # pragma: no cover
        try:
            out = json.loads(val)
        except ValueError:
            if val.startswith('[') and val.endswith(']'):
                if val[1:-1]:
                    out = [self.backwards_str2val(x.strip())
                           for x in val[1:-1].split(',')]
                else:
                    out = []
            elif val.startswith("'") and val.endswith("'"):
                out = val.strip("'")
            else:
                out = val
        return out

    def get(self, section, option, default=None, **kwargs):
        r"""Return None if the section/option does not exist.

        Args:
            section (str): Name of section.
            option (str): Name of option in section.
            default (obj, optional): Value that should be returned if the
                section and/or option are not found or are an empty string.
                Defaults to None.
            **kwargs: Additional keyword arguments are passed to the parent
                class's get.

        Returns:
            obj: String entry if the section & option exist, otherwise default.

        """
        section = section.lower()
        option = option.lower()
        if self.has_section(section) and self.has_option(section, option):
            # Super does not work for ConfigParser as not inherited from object
            out = configparser.ConfigParser.get(self, section, option, **kwargs)
            # Count empty strings as not provided
            if not out:
                return default
            else:
                return self.backwards_str2val(out)
        else:
            return default


def get_language_order(drivers):
    r"""Get the correct language order, including any base languages.

    Args:
        drivers (dict): Drivers in order.

    Returns:
        dict: Drivers in sorted order.

    """
    from yggdrasil.components import import_component
    out = OrderedDict()
    for d, drv in drivers.items():
        if d == 'cpp':
            d = 'c++'
        new_deps = OrderedDict()
        sub_deps = OrderedDict()
        for sub_d in drv.base_languages:
            if sub_d in drivers:
                sub_deps[sub_d] = drivers[sub_d]
            else:
                sub_deps[sub_d] = import_component('model', sub_d)
        sub_deps = get_language_order(sub_deps)
        min_dep = -1
        for sub_d, sub_drv in sub_deps.items():
            if sub_d in out:
                min_dep = max(min_dep, list(out.keys()).index(sub_d))
            else:
                new_deps[sub_d] = sub_drv
        if d in out.keys():
            dpos = list(out.keys()).index(d)
            assert(dpos > min_dep)
            min_dep = dpos
        else:
            new_deps[d] = drv
        new_out = OrderedDict()
        for k in list(out.keys())[:(min_dep + 1)]:
            new_out[k] = out[k]
        new_out.update(new_deps)
        for k in list(out.keys())[(min_dep + 1):]:
            new_out[k] = out[k]
        out = new_out
    return out


def update_language_config(languages=None, skip_warnings=False,
                           disable_languages=None, enable_languages=None,
                           lang_kwargs=None, overwrite=False, verbose=False):
    r"""Update configuration options for a language driver.

    Args:
        languages (list, optional): List of languages to configure.
            Defaults to None and all supported languages will be
            configured.
        skip_warnings (bool, optional): If True, warnings about missing options
            will not be raised. Defaults to False.
        disable_languages (list, optional): List of languages that should be
            disabled. Defaults to an empty list.
        enable_languages (list, optional): List of languages that should be
            enabled. Defaults to an empty list.
        overwrite (bool, optional): If True, the existing file will be overwritten.
            Defaults to False.
        verbose (bool, optional): If True, information about the config file
            will be displayed. Defaults to False.
        lang_kwargs (dict, optional): Dictionary containing language
            specific keyword arguments. Defaults to {}.

    """
    from yggdrasil.components import import_component
    if verbose:
        logger.info("Updating user configuration file for yggdrasil at:\n\t%s"
                    % usr_config_file)
    miss = []
    if (languages is None) or overwrite:
        all_languages = tools.get_supported_lang()
        languages = ['c', 'c++', 'make', 'cmake', 'python', 'lpy', 'r', 'matlab']
        for lang in all_languages:
            if lang.lower() not in languages:
                languages.append(lang)
    elif not isinstance(languages, list):
        languages = [languages]
    if disable_languages is None:
        disable_languages = []
    if enable_languages is None:
        enable_languages = []
    if lang_kwargs is None:
        lang_kwargs = {}
    if overwrite:
        shutil.copy(def_config_file, usr_config_file)
        ygg_cfg_usr.reload()
    drivers = OrderedDict([(lang, import_component('model', lang))
                           for lang in languages])
    drv = list(get_language_order(drivers).values())
    for idrv in drv:
        if (((idrv.language in disable_languages)
             and (idrv.language in enable_languages))):
            logger.info(("%s language both enabled and disabled. "
                         "No action will be taken.") % idrv.language)
        elif idrv.language in disable_languages:
            ygg_cfg_usr.set(idrv.language, 'disable', 'True')
        elif idrv.language in enable_languages:
            ygg_cfg_usr.set(idrv.language, 'disable', 'False')
        if ygg_cfg_usr.get(idrv.language, 'disable', 'False').lower() == 'true':
            continue  # pragma: no cover
        miss += idrv.configure(ygg_cfg_usr,
                               **lang_kwargs.get(idrv.language, {}))
    ygg_cfg_usr.update_file()
    ygg_cfg.reload()
    if not skip_warnings:
        for sect, opt, desc in miss:  # pragma: windows
            warnings.warn(("Could not set option %s in section %s. "
                           + "Please set this in %s to: %s")
                          % (opt, sect, ygg_cfg_usr.file_to_update, desc),
                          RuntimeWarning)
    

# Set associated environment variables
env_map = [('debug', 'ygg', 'YGG_DEBUG'),
           ('debug', 'rmq', 'RMQ_DEBUG'),
           ('debug', 'client', 'YGG_CLIENT_DEBUG'),
           ('jsonschema', 'validate_components', 'YGG_SKIP_COMPONENT_VALIDATION'),
           ('jsonschema', 'validate_all_messages', 'YGG_VALIDATE_ALL_MESSAGES'),
           ('rmq', 'namespace', 'YGG_NAMESPACE'),
           ('rmq', 'host', 'YGG_MSG_HOST'),
           ('rmq', 'vhost', 'YGG_MSG_VHOST'),
           ('rmq', 'user', 'YGG_MSG_USER'),
           ('rmq', 'password', 'YGG_MSG_PW'),
           ('parallel', 'cluster', 'YGG_CLUSTER'),
           ]


def get_ygg_loglevel(cfg=None, default='DEBUG'):
    r"""Get the current log level.

    Args:
        cfg (:class:`yggdrasil.config.YggConfigParser`, optional):
            Config parser with options that should be used to determine the
            log level. Defaults to :data:`yggdrasil.config.ygg_cfg`.
        default (str, optional): Log level that should be returned if the log
            level option is not set in cfg. Defaults to 'DEBUG'.

    Returns:
        str: Log level string.

    """
    is_model = tools.is_subprocess()
    if cfg is None:
        cfg = ygg_cfg
    if is_model:
        opt = 'client'
    else:
        opt = 'ygg'
    return cfg.get('debug', opt, default)


def set_ygg_loglevel(level, cfg=None):
    r"""Set the current log level.

    Args:
        level (str): Level that the log should be set to.
        cfg (:class:`yggdrasil.config.YggConfigParser`, optional):
            Config parser with options that should be used to update the
            environment. Defaults to :data:`yggdrasil.config.ygg_cfg`.
    
    """
    is_model = tools.is_subprocess()
    if cfg is None:
        cfg = ygg_cfg
    if is_model:
        opt = 'client'
    else:
        opt = 'ygg'
    cfg.set('debug', opt, level)
    logLevelYGG = eval('logging.%s' % level)
    ygg_logger = logging.getLogger("yggdrasil")
    ygg_logger.setLevel(level=logLevelYGG)
        
    
def cfg_logging(cfg=None):
    r"""Set logging levels from config options.

    Args:
        cfg (:class:`yggdrasil.config.YggConfigParser`, optional):
            Config parser with options that should be used to update the
            environment. Defaults to :data:`yggdrasil.config.ygg_cfg`.

    """
    is_model = tools.is_subprocess()
    if cfg is None:
        cfg = ygg_cfg
    _LOG_FORMAT = "%(levelname)s:%(module)s.%(funcName)s[%(lineno)d]:%(message)s"
    logging.basicConfig(level=logging.INFO, format=_LOG_FORMAT)
    logLevelYGG = eval('logging.%s' % cfg.get('debug', 'ygg', 'NOTSET'))
    logLevelRMQ = eval('logging.%s' % cfg.get('debug', 'rmq', 'INFO'))
    logLevelCLI = eval('logging.%s' % cfg.get('debug', 'client', 'INFO'))
    ygg_logger = logging.getLogger("yggdrasil")
    rmq_logger = logging.getLogger("pika")
    if is_model:
        ygg_logger.setLevel(level=logLevelCLI)
    else:
        ygg_logger.setLevel(level=logLevelYGG)
    rmq_logger.setLevel(level=logLevelRMQ)
    # For models, route the loggs to stdout so that they are displayed by the
    # model driver.
    if is_model:
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(logLevelCLI)
        ygg_logger.addHandler(handler)
        rmq_logger.addHandler(handler)


def cfg_environment(env=None, cfg=None):
    r"""Set environment variables based on config options.

    Args:
        env (dict, optional): Dictionary of environment variables that should
            be updated. Defaults to `os.environ`.
        cfg (:class:`yggdrasil.config.YggConfigParser`, optional):
            Config parser with options that should be used to update the
            environment. Defaults to :data:`yggdrasil.config.ygg_cfg`.

    """
    if env is None:
        env = os.environ
    if cfg is None:
        cfg = ygg_cfg
    for s, o, e in env_map:
        v = cfg.get(s, o)
        if v:
            env[e] = v

            
# Initialize config
ygg_cfg_usr = YggConfigParser.from_files([usr_config_file])
ygg_cfg = YggConfigParser.from_files([def_config_file, usr_config_file,
                                      loc_config_file])


# Do initial update of logging & environment (legacy)
cfg_logging()
cfg_environment()



from yggdrasil.tests import assert_raises
from yggdrasil.metaschema.datatypes.tests import test_MetaschemaType as parent


class TestAnyMetaschemaType(parent.TestMetaschemaType):
    r"""Test class for AnyMetaschemaType class."""

    _mod = 'AnyMetaschemaType'
    _cls = 'AnyMetaschemaType'
    
    @staticmethod
    def after_class_creation(cls):
        r"""Actions to be taken during class construction."""
        parent.TestMetaschemaType.after_class_creation(cls)
        cls._value = {'a': int(1), 'b': float(1)}
        cls._valid_encoded = [dict(cls._typedef,
                                   type=cls.get_import_cls().name,
                                   temptype={'type': 'int'})]
        cls._valid_decoded = [cls._value, object]
        cls._invalid_validate = []
        cls._invalid_encoded = [{}]
        cls._invalid_decoded = []
        cls._compatible_objects = [(cls._value, cls._value, None)]

    def test_decode_data_errors(self):
        r"""Test errors in decode_data."""
        assert_raises(ValueError, self.import_cls.decode_data, 'hello', None)



import pprint
from yggdrasil.tests import YggTestClass


class TestTransformBase(YggTestClass):
    r"""Test for TransformBase communication flass."""

    transform = 'TransformBase'
    skip_comm_check = True
    
    @property
    def cls(self):
        r"""str: Communication class."""
        return self.transform

    @property
    def mod(self):
        r"""str: Absolute module import."""
        return 'yggdrasil.communication.transforms.%s' % self.cls

    @property
    def inst_kwargs(self):
        r"""dict: Keyword arguments for tested class."""
        out = {}
        opt = self.get_options()
        if opt:
            out = opt[0].get('kwargs', {})
        return out

    def test_transform(self):
        r"""Test transform."""
        for x in self.get_options():
            inst = self.import_cls(**x.get('kwargs', {}))
            for msg_in, msg_exp in x.get('in/out', []):
                if isinstance(msg_exp, type(BaseException)):
                    self.assert_raises(msg_exp, inst, msg_in)
                else:
                    msg_out = inst(msg_in)
                    try:
                        self.assert_equal(msg_out, msg_exp)
                    except BaseException:  # pragma: debug
                        for t, x in zip(['in', 'expected', 'out'],
                                        [msg_in, msg_exp, msg_out]):
                            print('%s:' % t)
                            pprint.pprint(x)
                        raise

    def test_transform_type(self):
        r"""Test transform_type."""
        for x in self.get_options():
            inst = self.import_cls(**x.get('kwargs', {}))
            for typ_in, typ_exp in x.get('in/out_t', []):
                if isinstance(typ_exp, type(BaseException)):
                    self.assert_raises(typ_exp, inst.validate_datatype,
                                       typ_in)
                else:
                    inst.validate_datatype(typ_in)
                    typ_out = inst.transform_datatype(typ_in)
                    try:
                        self.assert_equal(typ_out, typ_exp)
                    except BaseException:  # pragma: debug
                        for t, x in zip(['in', 'expected', 'out'],
                                        [typ_in, typ_exp, typ_out]):
                            print('%s:' % t)
                            pprint.pprint(x)
                        raise

    def test_transform_empty(self):
        r"""Test transform of empty bytes message."""
        self.assert_equal(self.instance(b'', no_init=True), b'')
